package examples.jce.rsa_aes;

import java.io.*;
import javax.crypto.*;
import java.security.*;
import java.security.cert.*;
import javax.crypto.spec.SecretKeySpec;
import javax.crypto.spec.IvParameterSpec;


/**
 * Decrypt a file using HMAC for message integrity checking.
 * This class can be used from the command line
 * DecryptFileHMAC [Source Encrypted file] [Destination Decrypted File] [keystore]
 */


public class DecryptFileHMAC
{
   
    private PrivateKey   priv_key = null; // RSA decryption key.
    private SecureRandom sec_rand = null; // Class wide secure random.
    
    private File file_in = null; // Source File. 
    private File file_out = null; // Output File.

   


    /**
     * Main method
     */

    public static void main(String[] args)
    {
        new DecryptFileHMAC(args);
    }

    
  

    
    /**
     * Initialize file decryption type constructor with pre generated parameters.
     */

    public DecryptFileHMAC(
                           File in, 
                           File out, 
                           KeyStore key_store,
                           String alias, 
                           char[] ks_pass) throws Exception
    {
        init(in,out,key_store,alias,ks_pass);
    }   


    /**
     * Constructor for command line.
     */
      
    public DecryptFileHMAC(String[] args)
    {
        try {
            initTerm(args);
            decrypt();
        } catch (Exception ex)
            {
                ex.printStackTrace();
            }
    }



    /**
     * Initialize the class.
     */

    public void init(File file_in, 
                     File file_out, 
                     KeyStore key_store,
                     String alias, 
                     char[] ks_pass) throws Exception
    {
        
        //
        // Load provider and create SecureRandom instance.
        // Also salt the secure random instance from /dev/urandom if such a file exists.
        // '/dev/urandom' exists on most UNIX style machines.
        // 

        try
            {
                Security.addProvider(new org.bouncycastle.jce.provider.BouncyCastleProvider());
                sec_rand = SecureRandom.getInstance("SHA1PRNG","SUN");
                
                if (new File("/dev/urandom").exists())
                    {
                        System.out.println("Salting SecureRandom (SHA1PRNG) from /dev/urandom");
                        byte[] salt = new byte[8192];
                        new FileInputStream("/dev/urandom").read(salt);
                        sec_rand.setSeed(salt);
                    }
            }
        catch (Exception ex)
            {
                ex.printStackTrace();
            }

        this.file_in = file_in;
        this.file_out = file_out;
        priv_key = (PrivateKey)key_store.getKey(alias,ks_pass);
    }




    /**
     * initTerm manages the initialization from a command line.
     * It queries the user for the key store password and the alias of their private key
     */

    public void initTerm(String[] args)
    {
        
        //
        // Load provider and create SecureRandom instance.
        // Also salt the secure random instance from /dev/urandom if such a file exists.
        // '/dev/urandom' exists on most UNIX style machines.
        // 

        try
            {
                Security.addProvider(new org.bouncycastle.jce.provider.BouncyCastleProvider());
                sec_rand = SecureRandom.getInstance("SHA1PRNG","SUN");
                
                if (new File("/dev/urandom").exists())
                    {
                        System.out.println("Salting SecureRandom (SHA1PRNG) from /dev/urandom");
                        byte[] salt = new byte[8192];
                        new FileInputStream("/dev/urandom").read(salt);
                        sec_rand.setSeed(salt);
                    }
            }
        catch (Exception ex)
            {
                ex.printStackTrace();
            }
        
        //
        // Check for the minimum number of entries
        //

        if (args.length <3)
        {
            System.out.println("Usage: <Source Encrypted file> <Destination Decrypted File> <key store>");
            System.exit(0);
        }

        file_in = new File(args[0]);  // Source file.
        file_out = new File(args[1]); // Destination file. (For plain text)

        if (!file_in.exists())
        {
            System.out.println("Source File does not exist..!");
            System.exit(0);
        }
    
        File keystore_file = new File(args[2]); // File for key store.

        //
        // Check the file exists.
        //

        if (!keystore_file.exists())
        {
            System.out.println("Key store file specified could not be found.");
            System.exit(0);
        }


        //
        // Now we need to set up the key store so that we can access the
        // private key.
        // For this we need to ask the user for the key store password and the alias the private
        // key is stored under.
        //


        char[]      ks_pass = null; // Keystore password.
        boolean     stop = false; // Stop flag.
        KeyStore    key_store = null; // The keystore.
       
        try 
        {
            key_store = KeyStore.getInstance("JKS","SUN"); // Get an instance of our key store.

            while(!stop)
            {
                System.out.println("");
                System.out.print("Key Store Password > ");
                String l = getLine(); // Call our method to read from stdin.
                
                //
                // A little sanity checking.
                //

                if (l == null)
                {
                    System.exit(0);
                }
                
                if (l.length() == 0)
                {
                    System.exit(0);
                }

                ks_pass = l.toCharArray(); // To char array.

                //
                // Test it against out key store.
                // The KeyStore will through an exception if the password is incorrect.
                //

                try
                {
                    key_store.load(new FileInputStream(keystore_file),ks_pass);
                    stop=true; // it worked so call stop.
                }
                catch (Exception ex)
                {
                    System.out.println(ex.getMessage());
                }
            }
        

            //
            // Reset the stop flag so we can ask for an alias.
            //

            stop = false;
        
            while(!stop)
            {
                System.out.println("");
                System.out.print("Secret Key Alias > ");
                String l = getLine();
                if (l == null)
                {
                    System.exit(0);
                }
                
                if (l.length() == 0)
                {
                    System.exit(0);
                }

                //
                // At this stage we can fetch the private key from the key store if
                // the alias is correct.
                // It will throw and exception if it is incorrect.
                //
                
                try
                {
                    priv_key = (PrivateKey)key_store.getKey(l,ks_pass);
                    stop = true; // It worked so flag a stop.
                }
                catch (Exception ex)
                {
                    System.out.println(ex.getMessage());
                }
            }
       
        } catch (Exception ex)
            {
                ex.printStackTrace();
            }
    }





    /**
     * The decrypt method sequences the decryption process by calling
     * parseHeader(), validateHMAC(), decryptFile() in that order.
     */

    public void decrypt() throws Exception
    {
             
        DecryptHMAC dec_hmac = new DecryptHMAC();

        //
        // Now we have the private key we can begin to decrypt the content of the
        // files header that are encrypted with RSA.
        // So we call out parseHeader() method with the appropriate information.
        //
        
        dec_hmac.parseHeader(file_in, priv_key, sec_rand);
        
        //
        // The parseHeader() method has loaded 'mac_key', 'aes_key', 'aes_iv', 'mac_code' with their
        // respective values. These variables are within the scope of the class.
        //
        
        //
        // Call validateHMAC() to test the integrity of the file.
        //
        
        if (dec_hmac.validateHMAC(file_in))
            {
                // The integrity of the file has been assured so
                // we can call the decryptFile() method and save the decrypted file
                // to 'file_out'.
                
                dec_hmac.decryptFile(file_in, file_out);
            }
        else
            {
                System.out.println("HMAC failed to validate");
            }   
    }



    /**
     * Read a line from System.in
     */
    private String getLine()
    {
        try
        {
            BufferedReader bin = new BufferedReader(new InputStreamReader(System.in));
            return(bin.readLine());
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
        return(null);
    }


}







