package examples.jce.rsa_aes;

import java.io.*;
import javax.crypto.*;
import java.security.*;
import java.security.cert.*;
import javax.crypto.spec.SecretKeySpec;
import javax.crypto.spec.IvParameterSpec;



/**
 * Decrypt a file and use signatures MD5 with RSA for message validation and authentication.
 * This class can be used as a command line tool..
 * DecryptFileSIG [Source Encrypted file] [Destination Decrypted File] [key store]
 */




public class DecryptFileSIG
{
  

    private File file_in = null;
    private File file_out = null;
    private PrivateKey priv_key = null;
    private SecureRandom sec_rand = null;


    final int           FILE_HEADER = 0x7e01; // File header.
    final int           DATA_BLOCK = 1; // Header for Data block.
    final int           FINAL_DATA_BLOCK = 2; // Header for last data block.
    final int           SIG_BLOCK = 3; // Header for Signature Block.
    final int           CERT_BLOCK = 4; // Header for senders certificate block.

    final int           KEY_BLOCK = 16; // Header for key block. (AES)
    final int           IV_BLOCK = 17; // Header for IV block. (AES)
    final int           LOCK_BLOCK = 18; // Header for locking block.
    
    /**
     * zero out the passed in byte array
     */
    private static void blank(
        byte[]   bytes)
    {
        for (int t = 0; t < bytes.length; t++)
        {
            bytes[t]=0;
        }
    }

    /**
     * Main Method.
     */

    public static void main(String[] args)
    {
        new DecryptFileSIG(args);
    }



    /**
     * Constructor for Terminal reading.
     */

    public DecryptFileSIG(String[] args)
    {
        try {
            initTerm(args);
            decryptFile();
        } catch (Exception ex)
            {
                ex.printStackTrace();
            }
    }



    /**
     * Blank constructor
     */

    public DecryptFileSIG()
    {

    }


    /**
     * Constructor with parameters
     */

    public DecryptFileSIG(File file_in, File file_out, PrivateKey priv_key,SecureRandom sec_rand)
    {
        init(file_in, file_out, priv_key,sec_rand);
    }



    /**
     * Initialise
     */

    public void init(File file_in, File file_out, PrivateKey priv_key,SecureRandom sec_rand)
    {

        this.sec_rand = null;
        this.file_in = file_in;
        this.file_out = file_out;
        this.priv_key = priv_key;
    }




    /**
     * Initialise by using command line arguments and asking user for input
     */

    public void initTerm(String[] args)
    {
        System.out.println("Decrypt file that uses Signature (MD5withRSA) to provide");
        System.out.println("to provide sender authentication and message integrity checking");

        sec_rand = null; // A secure random number generator.

        //
        // Add provider and seed SecureRandom with salt from '/dev/urandom' if device exists
        //

        try
        {
            Security.addProvider(new org.bouncycastle.jce.provider.BouncyCastleProvider());
            sec_rand = SecureRandom.getInstance("SHA1PRNG","SUN"); // SHA bases prng.
            
            if (new File("/dev/urandom").exists())
            {
                System.out.println("Salting SecureRandom (SHA1PRNG) from /dev/urandom");
                byte[] salt = new byte[8192];
                new FileInputStream("/dev/urandom").read(salt);
                sec_rand.setSeed(salt);
            }

        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }

        //
        // Test for minimum argument count. 
        //

        if (args.length <3)
        {
            System.out.println("Usage: <Source Encrypted file> <Destination Decrypted File> <keystore>");
            System.exit(0);
        }

        file_in = new File(args[0]); // File in. (cipher text)
        file_out = new File(args[1]); // File out. (plaintex)


        //
        // Test for source file.
        //

        if (!file_in.exists())
        {
            System.out.println("Source File does not exist..!");
            System.exit(0);
        }
    
        
        File keystore_file = new File(args[2]); // Keystore file.

        //
        // Test for existence of KeyStore file.
        //

        if (!keystore_file.exists())
        {
            System.out.println("Keystore file specified could not be found.");
            System.exit(0);
        }


        char[]          ks_pass = null; // Keystore password.
        boolean         stop = false; // Reused boolean for controlling loops.
        KeyStore        key_store = null; // The key store.
    
        try
        {
            key_store = KeyStore.getInstance("JKS","SUN");  // Instantiate key store.       


            //
            // Loop here until user inputs a valid password.
            // Break loop and exit program if user hits enter.
            //

            while(!stop)
            {
                System.out.println("");
                System.out.print("My Key Store Password > ");
                String l = getLine(); // Read in line from System.in.
                
                //
                // Some basic checks.
                //

                if (l == null)
                {
                    System.exit(0);
                }
                
                if (l.length() == 0)
                {
                    System.exit(0);
                }

                ks_pass = l.toCharArray(); // input to char[]

                //
                // Test password on keystore.
                // KeyStore will throw an exception if password incorrect.
                //
                try
                {
                    key_store.load(new FileInputStream(keystore_file), ks_pass); // load in keystore file.
                    stop=true; // Password correct so break loop.
                }
                catch (Exception ex)
                {
                    System.out.println(ex.getMessage());
                }
            }
        
            stop = false; // Reset loop breaker.
        

            //
            // We now need to obtain from the user the alias for
            // their secret key.
            //

            while(!stop)
            {
                System.out.println("");
                System.out.print("Decryption Secret Key Alias  > ");
                String l = getLine();

                // Some sanity checks.

                if (l == null)
                {
                    System.exit(0);
                }
                
                if (l.length() == 0)
                {
                    System.exit(0);
                }
                
                //
                // Try fetch the secret key for a given alias.
                // KeyStore will throw an exception in alias not present.
                //

                try
                {
                    priv_key = (PrivateKey)key_store.getKey(l, ks_pass); // Fetch private key.
                    stop=true; // It worked so flag a break.
                }
                catch (Exception ex)
                {
                    System.out.println(ex.getMessage());
                }
            } // loop.
        
        } catch (Exception ex)
            {
                ex.printStackTrace();
            }

   
    }


        /**
         * Method to get line from terminal (System.in);
         */


    private String getLine()
    {
        try
        {
            BufferedReader bin = new BufferedReader(new InputStreamReader(System.in));
            return(bin.readLine());
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }

        return(null);
    }



    /**
     * Decrypt file.
     */

    public void decryptFile() throws Exception
        {
            //
            // Call parse header to decrypt and extract information
            // from the header.
            // This method will extract and define, 'lock_data', 'their_cert', 'aes_key', 'aes_iv', 'sig_code'
            //
            DecryptSIG dec_sig = new DecryptSIG();
            
            dec_sig.parseHeader(file_in, sec_rand, priv_key);
            
            //
            // After parsing the file you should take the senders certificate
            // and validate that against your trust store or what ever mechanism
            // you are using to authenticate the sender of the message.
            //
            // If your system cannot validate the certificate then drop the
            // message. People cause themselves a lot of problems by allowing
            // users to ignore a message that the system can not validate.
            //
            // You must not allow people to bypass or ignore your
            // authentication system. If you do you may as well not have a
            // system of authentication in the first place!
            //
            // This can be a lot more complicated than it looks and, at any
            // rate, it is beyond the scope of the example because I would
            // have to build a whole authentication frame work to get my point
            // across.
            //
            // X509Certificate their_Cert = dec_sig.getSendersCert();

            if (dec_sig.validateSIG(file_in)) // Check the file's integrity.
            {
                dec_sig.decryptFile(file_in, sec_rand, file_out); // Decrypt the file.
            }
            else // Didn't validate.
            {
                System.out.println("SIG failed to validate");
            }           
        
        }



    // Have fun..
}





