package examples.jce.session;

import java.io.*;
import java.net.*;
import java.util.*;

import javax.crypto.*;
import java.security.*;
import java.security.spec.*;
import java.security.cert.*;

import javax.crypto.spec.*;

/**
 * <p>
 * NOTE: 
 * Session establishment is easier to understand from the clients perspective.
 * The ClientSide.java source code has a lot of comments and a blow by blow
 * discussion of the session establishment.
 * This is the servers' side of the session establishment.
 * The server process follows the client process and will time out after 10 seconds of inactivity.
 */

public class ServerSide extends Thread
{
    private KeyStore        key_store = null;
    private char[]          ks_pass = null;

    private int             port = 1887;
    private SecureRandom    sec_rand = null;
    private boolean         socket_ready = false;

    private Thread          server_thread = null;
    private boolean         server_thread_stop = false;

    /**
     * zero out the passed in byte array
     */
    private static void blank(
        byte[]   bytes)
    {
        for (int t = 0; t < bytes.length; t++)
        {
            bytes[t]=0;
        }
    }

    public ServerSide(
        KeyStore key_store, 
        char[] ks_pass, 
        int port, 
        SecureRandom sec_rand)
    {
        this.key_store = key_store;
        this.ks_pass = ks_pass;
        this.port = port;
        this.sec_rand = sec_rand;
    }

    /**
     * This method starts the server thread and then
     * blocks until the ServerSocket is set up and ready
     * for connections.
     */
    public void invokeServer()
    {
        start();
        while(!socket_ready)
        {
            // loop...
        }
    }

    public void run()
    {
        try
        {
            byte[] se_rand = new byte[32]; // Server Random.
            byte[] cl_rand = new byte[32]; // Client Random.
            
            Cipher          cl_cipher = Cipher.getInstance("AES/CBC/PKCS7Padding","BC"); // Client Cipher.
            Cipher          se_cipher = Cipher.getInstance("AES/CBC/PKCS7Padding","BC"); // Server Cipher.
            Cipher          rsa_eng = Cipher.getInstance("RSA/None/OAEPPadding","BC"); // RSA Cipher.
            Mac             cl_mac = Mac.getInstance("HMACSHA1","BC"); // Client side MAC
            Mac             se_mac = Mac.getInstance("HMACSHA1","BC"); // Server side MAC
            PublicKey       cl_pubkey = null; // The clients public key for encryption and decryption.
            X509Certificate cl_cert = null; // The clients authentication and signing certificate.
            
            //
            // Create and setup a server socket to listen to
            // incoming connections.
            //
            ServerSocket    s_sock = new ServerSocket(port);
            Socket          sock = null;

            s_sock.setSoTimeout(2000);

            for (;;)
            {
                socket_ready = true;
            
                try
                {
                    sock = s_sock.accept();
                }
                catch (InterruptedIOException iioex)
                {
                    continue;
                }

                sock.setSoTimeout(10000);
        
                //
                // Connection accepted
                // 

                //
                // Set up stream wrappers.
                //
                
                ExDataInputStream   data_in = new ExDataInputStream(sock.getInputStream());
                ExDataOutputStream  data_out = new ExDataOutputStream(sock.getOutputStream());
            
                //
                // Step one..
                // Upon receipt of client Hello the server process reads
                // in a byte array of 32 random bytes. The server process
                // then responds with a byte array of random data.
                //
                
                if (data_in.readShort() == SessionConstants.HELLO)
                {
                    System.out.println("Server: got HELLO from client");
                    cl_rand = new byte[32];
                    data_in.readFully(cl_rand);
                
                    // 
                    // Generate Server's Random;
                    // 
                    sec_rand.nextBytes(se_rand); // Fill with random data.
                    data_out.writeShort(SessionConstants.HELLO);
                    data_out.write(se_rand); // Write out random.
                    data_out.flush(); // Flush stream.      
                    System.out.println("Server: Sending my HELLO to client");
                }
                else
                {
                    throw new Exception("Session broken at HELLO stage");
                }

                
                //
                // Step two..
                // 1. The server reads in the clients public key.
                // 2. The clients signing certificate.
                // 3. A signature.
                //
                // Note that the signature is composed of:
                //    Client public key;
                //    Concatenation of Clients random and Server random;
                //    Clients signing certificate.
                //
                // By including the shared randoms we effectively make
                // this exchange unique each time a session is started.
                // This would stop an attacker from replaying a previously
                // recorded exchange.  
                //

                if (data_in.readShort() == SessionConstants.PUBLIC_KEY_EXCHANGE)
                {                       
                    System.out.println("Server: Reading public key + cert + sig from client");
                    byte[]              pub_key = data_in.readByteArray(); // Read public key

                    //
                    // Now we must take in the encoded public key and
                    // use a KeyFactory to turn it back into a Key Object.
                    //

                    X509EncodedKeySpec  key_spec = new X509EncodedKeySpec(pub_key);
                    KeyFactory          key_fact= KeyFactory.getInstance("RSA", "BC");
                    cl_pubkey = key_fact.generatePublic(key_spec); // Generate the public key.

                    //
                    // Read in encoded Certificate and use a CertificateFactory
                    // to convert it into a X509Certificate
                    //
                    byte[]              cert = data_in.readByteArray(); // Read in cert.
                    CertificateFactory  cf = CertificateFactory.getInstance("X.509");
                    cl_cert = (X509Certificate) cf.generateCertificate(new ByteArrayInputStream(cert)); // Convert.
                
        	
                    //
                    // Now we must read in the signature code and use that
                    // to verify the integrity of the exchanged public key.
                    // NOTE:
                    // At this stage the server process should attempt to
                    // validate the certificate against the authentication
                    // model that your system is using.
                    // 
                    byte[]              sig_code = data_in.readByteArray();             // Read signature.
                    Signature           sig = Signature.getInstance("MD5withRSA","BC"); // Set up Signer.

                    sig.initVerify(cl_cert);    // Initialize with clients signing certificate.
                    sig.update(cl_rand);        // Incorporate client random.
                    sig.update(se_rand);        // Incorporate server random.
                    sig.update(pub_key);        // Incorporate encoded public key.
                    sig.update(cert);           // Incorporate encoded certificate.
                
                    //
                    // Test signature and fail if it does not verify.
                    //
                    if (!sig.verify(sig_code))
                    {
                        throw new Exception("Signature failed on Public key exchange data unit");
                    }
                    else
                    {
                       System.out.println("Server: Client PDU signature passed");
                    }
                    
                    System.out.println("Server: Sending my public key + cert + sig to client.");
                
                    //
                    // Send server public key to client.
                    //

                    //
                    // Set up signer.
                    //
                    sig.initSign((PrivateKey)key_store.getKey("hostsig",ks_pass));
                    sig.update(cl_rand);
                    sig.update(se_rand);
                
                    //
                    // Get my public key (for encryption) as a byte array.
                    //
                    byte[] my_pub = ((X509Certificate)key_store.getCertificate("hostenc")).getPublicKey().getEncoded();

                    //
                    // Get my certificate (for sig validation and auth) as
                    // a byte array.
                    //
                    byte[] my_cert = ((X509Certificate)key_store.getCertificate("hostsig")).getEncoded();
                
                    sig.update(my_pub); 
                    sig.update(my_cert); 
                
                    data_out.writeShort(SessionConstants.PUBLIC_KEY_EXCHANGE);    
                    data_out.writeByteArray(my_pub);
                    data_out.writeByteArray(my_cert);
                
                    sig_code = sig.sign();
                
                    data_out.writeByteArray(sig_code);
                
                    data_out.flush();
                }

                //
                // Part 3: The Secret Exchange..
                // 
                // 1. Read SecretKey encrypted with RSA.
                // 2. Read IV encrypted with RSA.
                // 3. Read HMAC key encrypted with RSA.
                // 4. Read in signature.
                // 
                // NOTE: No need to send cert again. Certificate
                // sent in previous stage.
                //
                if (data_in.readShort() == SessionConstants.SECRET_EXCHANGE)
                {
                    Cipher aes_lock = Cipher.getInstance("AES/CBC/PKCS7Padding");
    
                    rsa_eng.init(
                         Cipher.DECRYPT_MODE,
                         (PrivateKey)key_store.getKey("hostenc",ks_pass),
                         sec_rand);
                
                    //
                    // Read in secret key .
                    //
                    byte[] aes_key_enc = data_in.readByteArray();
                
                    // 
                    // Read in IV
                    //
                    byte[] iv_enc = data_in.readByteArray();
                
                    // 
                    // Read in HMAC key.
                    //
                    byte[] hmac_key_enc = data_in.readByteArray();
                
                    //
                    // Read in lock
                    //
                    byte[] tmp_lock = data_in.readByteArray();    
                
                    //
                    // Validate message against client's signing certificate
                    // exchanged in the previous stage.
                    //
                    // But first we need to decrypt the lock data to incorperate into the exchange.
                    //

                    aes_lock.init(Cipher.DECRYPT_MODE,
                           (SecretKey)new SecretKeySpec(rsa_eng.doFinal(aes_key_enc), "AES"),
                           new IvParameterSpec(rsa_eng.doFinal(iv_enc)));


                    Signature sig = Signature.getInstance("MD5withRSA", "BC");
                    sig.initVerify(cl_cert);
                    
                    sig.update(cl_rand); 
                    sig.update(se_rand);
                    sig.update(aes_key_enc);
                    sig.update(iv_enc);
                    sig.update(hmac_key_enc);
                    sig.update(aes_lock.doFinal(tmp_lock));
                
                    if (!sig.verify(data_in.readByteArray()))
                    {
                        throw new Exception("Signature failed on Public key exchange data unit");
                    }
                    else
                    {
                        System.out.println("Server: Client PDU for secret key exchange signature passed");
                    }

                    //
                    // Now we can set up a Cipher instance that will decrypt
                    // data sent from the client.
                    // 
                    cl_cipher.init(Cipher.DECRYPT_MODE,
                           (SecretKey)new SecretKeySpec(rsa_eng.doFinal(aes_key_enc), "AES"),
                           new IvParameterSpec(rsa_eng.doFinal(iv_enc)));
                
                    //
                    // Set up the MAC to validate data sent from the client
                    // side.
                    //
                    cl_mac.init((SecretKey)new SecretKeySpec(rsa_eng.doFinal(hmac_key_enc), "AES"));

                    //
                    // Now send my secrets back to client encrypted with
                    // public key exchanged by client.
                    //

                    //
                    // Generate my secrets.
                    //
                    KeyGenerator    key_gen = KeyGenerator.getInstance("AES","BC");

                    key_gen.init(192, sec_rand);

                    SecretKey       my_aes_key = key_gen.generateKey();
                
                    key_gen.init(160, sec_rand);

                    SecretKey       my_hmac_key = key_gen.generateKey();
                
                    se_cipher.init(Cipher.ENCRYPT_MODE,my_aes_key,sec_rand);

                    byte[] my_iv = se_cipher.getIV();
                
                    se_mac.init(my_hmac_key); 
                    tmp_lock = new byte[24];
                    sec_rand.nextBytes(tmp_lock);
                
                    //
                    // Set up RSA for encryption.
                    //
                    sig.initSign((PrivateKey) key_store.getKey("hostsig", ks_pass));
                    sig.update(cl_rand);
                    sig.update(se_rand);

                    rsa_eng.init(Cipher.ENCRYPT_MODE, cl_pubkey, sec_rand);
                    data_out.writeShort(SessionConstants.SECRET_EXCHANGE);
                
                    //
                    // Encrypt and send AES key.
                    //
                    byte[] tmp = rsa_eng.doFinal(my_aes_key.getEncoded());
                    data_out.writeByteArray(tmp);
                    sig.update(tmp);    
                    blank(tmp);
                
                    //
                    // Encrypt and send IV for cipher.
                    //
                    tmp = rsa_eng.doFinal(my_iv);
                    data_out.writeByteArray(tmp);
                    sig.update(tmp);    
                    blank(tmp);
                
                    //
                    // Encrypt and send MAC key..
                    //
                

                    tmp = rsa_eng.doFinal(my_hmac_key.getEncoded());
                    data_out.writeByteArray(tmp);
                    sig.update(tmp);    
                    blank(tmp);
                
                    //
                    // Encrypt and send LOCK data..
                    //
                    aes_lock.init(Cipher.ENCRYPT_MODE, my_aes_key, new IvParameterSpec(my_iv), sec_rand);
                    tmp = aes_lock.doFinal(tmp_lock);
                    data_out.writeByteArray(tmp);
                    sig.update(tmp_lock);            // Incorporate plain text into signature.
                    blank(tmp);
                

                    //
                    // Finalize and Send signature.
                    //
                    data_out.writeByteArray(sig.sign());
                    data_out.flush();


                    //
                    // Now we have finished exchanging symmetric keys we
                    // can set up our PDU generators to send encrypted
                    // messages.
                    //
                    data_out.initPDU(se_cipher, se_mac);
                    data_in.initPDU(cl_cipher, cl_mac);
    
                    System.out.println("Client -> Server: " + new String(data_in.readPDU(),"ISO-8859-1"));               
                    data_out.writePDU("I am fine.. Good to see the session working for you as well".getBytes("ISO-8859-1"));

                    System.out.println("Client -> Server: " + new String(data_in.readPDU(),"ISO-8859-1"));               
                    data_out.writePDU("Lift off.".getBytes("ISO-8859-1"));

                    System.out.println("Server exiting");
                    break;
                }
            }                
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }
}






