package examples.jce.rsa_aes;

import java.io.*;
import javax.crypto.*;
import java.security.*;
import java.security.cert.*;
import javax.crypto.spec.SecretKeySpec;
import javax.crypto.spec.IvParameterSpec;

public class DecryptFileSIG
{
    private byte[]          lock_data = null; // Lock data
    private X509Certificate their_cert = null; // Senders Authentication Certificate.
    private SecretKey       aes_key = null; // AES key.
    private byte[]          aes_iv = null;   // AES IV
    private byte[]          sig_code = null; // Signature.

    final int           FILE_HEADER = 0x7e01; // File header.
    final int           DATA_BLOCK = 1; // Header for Data block.
    final int           FINAL_DATA_BLOCK = 2; // Header for last data block.
    final int           SIG_BLOCK = 3; // Header for Signature Block.
    final int           CERT_BLOCK = 4; // Header for senders certificate block.

    final int           KEY_BLOCK = 16; // Header for key block. (AES)
    final int           IV_BLOCK = 17; // Header for IV block. (AES)
    final int           LOCK_BLOCK = 18; // Header for locking block.
    
    /**
     * zero out the passed in byte array
     */
    private static void blank(
        byte[]   bytes)
    {
        for (int t = 0; t < bytes.length; t++)
        {
            bytes[t]=0;
        }
    }

    /**
     * Main Method.
     */

    public static void main(String[] args)
    {
        new DecryptFileSIG(args);
    }



    /**
     * Constructor:
     * Takes in argument and decrypts file if arguments are sane.
     */ 
    
    public DecryptFileSIG(String[] args)
    {
        System.out.println("Decrypt file that uses Signature (MD5withRSA) to provide");
        System.out.println("to provide sender authentication and message integrity checking");

        SecureRandom    sec_rand = null; // A secure random number generator.

        //
        // Add provider and seed SecureRandom with salt from '/dev/urandom' if device exists
        //

        try
        {
            Security.addProvider(new org.bouncycastle.jce.provider.BouncyCastleProvider());
            sec_rand = SecureRandom.getInstance("SHA1PRNG","SUN"); // SHA bases prng.
            
            if (new File("/dev/urandom").exists())
            {
                System.out.println("Salting SecureRandom (SHA1PRNG) from /dev/urandom");
                byte[] salt = new byte[8192];
                new FileInputStream("/dev/urandom").read(salt);
                sec_rand.setSeed(salt);
            }

        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }

        //
        // Test for minimum argument count. 
        //

        if (args.length <3)
        {
            System.out.println("Usage: <Source Encrypted file> <Destination Decrypted File> <keystore>");
            System.exit(0);
        }

        File    file_in = new File(args[0]); // File in. (cipher text)
        File    file_out = new File(args[1]); // File out. (plaintex)


        //
        // Test for source file.
        //

        if (!file_in.exists())
        {
            System.out.println("Source File does not exist..!");
            System.exit(0);
        }
    
        
        File keystore_file = new File(args[2]); // Keystore file.

        //
        // Test for existence of KeyStore file.
        //

        if (!keystore_file.exists())
        {
            System.out.println("Keystore file specified could not be found.");
            System.exit(0);
        }


        char[]          ks_pass = null; // Keystore password.
        boolean         stop = false; // Reused boolean for controlling loops.
        KeyStore        key_store = null; // The key store.
        PrivateKey      priv_key = null; // My private key.

        try
        {
            key_store = KeyStore.getInstance("JKS","SUN");  // Instantiate key store.       


            //
            // Loop here until user inputs a valid password.
            // Break loop and exit program if user hits enter.
            //

            while(!stop)
            {
                System.out.println("");
                System.out.print("My Key Store Password > ");
                String l = getLine(); // Read in line from System.in.
                
                //
                // Some basic checks.
                //

                if (l == null)
                {
                    System.exit(0);
                }
                
                if (l.length() == 0)
                {
                    System.exit(0);
                }

                ks_pass = l.toCharArray(); // input to char[]

                //
                // Test password on keystore.
                // KeyStore will throw an exception if password incorrect.
                //
                try
                {
                    key_store.load(new FileInputStream(keystore_file), ks_pass); // load in keystore file.
                    stop=true; // Password correct so break loop.
                }
                catch (Exception ex)
                {
                    System.out.println(ex.getMessage());
                }
            }
        
            stop = false; // Reset loop breaker.
        

            //
            // We now need to obtain from the user the alias for
            // their secret key.
            //

            while(!stop)
            {
                System.out.println("");
                System.out.print("Decryption Secret Key Alias  > ");
                String l = getLine();

                // Some sanity checks.

                if (l == null)
                {
                    System.exit(0);
                }
                
                if (l.length() == 0)
                {
                    System.exit(0);
                }
                
                //
                // Try fetch the secret key for a given alias.
                // KeyStore will throw an exception in alias not present.
                //

                try
                {
                    priv_key = (PrivateKey)key_store.getKey(l, ks_pass); // Fetch private key.
                    stop=true; // It worked so flag a break.
                }
                catch (Exception ex)
                {
                    System.out.println(ex.getMessage());
                }
            } // loop.
        

            //
            // Call parse header to decrypt and extract information
            // from the header.
            // This method will extract and define, 'lock_data', 'their_cert', 'aes_key', 'aes_iv', 'sig_code'
            //

            parseHeader(file_in, sec_rand, priv_key);
        
            //
            // After parsing the file you should take the senders certificate
            // and validate that against your trust store or what ever mechanism
            // you are using to authenticate the sender of the message.
            //
            // If your system cannot validate the certificate then drop the
            // message. People cause themselves a lot of problems by allowing
            // users to ignore a message that the system can not validate.
            //
            // You must not allow people to bypass or ignore your
            // authentication system. If you do you may as well not have a
            // system of authentication in the first place!
            //
            // This can be a lot more complicated than it looks and, at any
            // rate, it is beyond the scope of the example because I would
            // have to build a whole authentication frame work to get my point
            // across.
            //

            if (validateSIG(file_in)) // Check the file's integrity.
            {
                decryptFile(file_in, sec_rand, file_out); // Decrypt the file.
            }
            else // Didn't validate.
            {
                System.out.println("SIG failed to validate");
            }           
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }




    private String getLine()
    {
        try
        {
            BufferedReader bin = new BufferedReader(new InputStreamReader(System.in));
            return(bin.readLine());
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }

        return(null);
    }




    public void decryptFile(
        File            file_in,
        SecureRandom    sec_rand,
        File            file_out)
        throws Exception
    {
        Cipher              cipher = Cipher.getInstance("AES/CBC/PKCS7Padding", "BC"); // Get Cipher object.

        //
        // Initialize Cipher object with SecretKey and IV for the Symmetric cipher (AES)
        //
        cipher.init(Cipher.DECRYPT_MODE, aes_key, new IvParameterSpec(aes_iv));
    

        FileOutputStream    file_str = new FileOutputStream(file_out); // Where to save plain text.
        
        // 
        // Input source file, via DataInputStream wrapper.
        //

        DataInputStream     data_str = new DataInputStream(new FileInputStream(file_in));


        boolean             stop = false; // Loop breaker.
        int                 cmd = 0; // Variable to hold block header.
        int                 l = 0; // Universal length variable.
        byte[]              buf = new byte[8192]; // A buffer to work in.

        for (;;)
        {
            cmd = data_str.readShort(); // Read in block header.

            if (cmd == DATA_BLOCK)
            {
                l = data_str.readInt(); // Get length of data.
                data_str.readFully(buf,0,l); // Read data.
                file_str.write(cipher.update(buf,0,l)); // Decrypt and save block.
            }
        
            if (cmd == FINAL_DATA_BLOCK)
            {
                l = data_str.readInt(); // Length of data.
                data_str.readFully(buf,0,l); // Read in data.
                file_str.write(cipher.doFinal(buf,0,l)); // Decrypt and save final block. (deal with padding etc)
                break;
            }


            //
            // The following blocks and their content are skipped.
            //

            if (cmd == KEY_BLOCK) // Skip the KEY Block
                {
                    l = data_str.readInt();
                    data_str.skip(l);
                }

            if (cmd == IV_BLOCK) // Skip the IV block.
                {
                    l = data_str.readInt();
                    data_str.skip(l);
                }
            
            if (cmd == LOCK_BLOCK) // Skip the IV block.
                {
                    l = data_str.readInt();
                    data_str.skip(l);
                }

            if (cmd == CERT_BLOCK) // Skip the IV block.
                {
                    l = data_str.readInt();
                    data_str.skip(l);
                }
            
            if (cmd == SIG_BLOCK)
                {
                    l = data_str.readInt();
                    data_str.skip(l);
                }
           
        }

        blank(buf); // Erase buffer.

        buf = null; // Set null.

        //
        // Flush and streams where necessary.
        //

        file_str.flush();
        file_str.close();
        data_str.close();
    }




    /**
     * This method, tests the validity of the file by using the Signature.
     */


    public boolean validateSIG(
        File    file_in)
        throws Exception
    {
        Signature               sig = Signature.getInstance("MD5withRSA","BC"); // Set up signature object.

        sig.initVerify(their_cert); // Initialize with senders certificate.
        sig.update(lock_data); // Update with lock data.
        
        //
        // Set up input stream wrappers.
        //
        
        SignatureInputStream    sig_str = new SignatureInputStream(new FileInputStream(file_in),sig);
        DataInputStream         data_str = new DataInputStream(sig_str);

        int cmd = 0; // variable to store block header.
        byte[] buf = new byte[8192]; // Buffer to work in.
        int l =0; // Universal length variable.

        //
        // It is worth while noting that I have used dummy reads instead of skipping
        // through the data. SignatureInputStream only processes data when a read method
        // is called. Now while the skip command my simply read through (n) bytes of data
        // on some streams it may actually do a skip which means that the data will not get
        // processed through the signature calculation.
        //

        for (;;)
        {
            cmd = data_str.readShort(); // Read off block header/

            if (cmd == DATA_BLOCK) // Skip HMAC block.
                {
                    l = data_str.readInt(); // Read length.
                    data_str.read(buf,0,l); // dummy read.
                }
            
            if (cmd == FINAL_DATA_BLOCK) // Skip the KEY Block
                {
                    l = data_str.readInt();
                    data_str.read(buf,0,l); // dummy read
                }


            if (cmd == KEY_BLOCK) // Skip the KEY Block
                {
                    l = data_str.readInt();
                    data_str.read(buf,0,l); // dummy read
                }

            if (cmd == IV_BLOCK) // Skip the IV block.
                {
                    l = data_str.readInt();
                    data_str.read(buf,0,l); // dummy read
                }
            
            if (cmd == LOCK_BLOCK) // Skip the IV block.
                {
                    l = data_str.readInt();
                    data_str.read(buf,0,l); // dummy read
                }

            if (cmd == CERT_BLOCK) // Skip the IV block.
                {
                    l = data_str.readInt();
                    data_str.read(buf,0,l);// dummy read
                }

            if (cmd == SIG_BLOCK)
            {
                break;
            }
        }
    
        data_str.close();
        
        //
        // Note that the sig code has already been extracted from the file by the
        // parseHeader() method.
        //

        return sig.verify(sig_code);
    }



    /**
     * The parseHeader() method extracts and decrypts the information contained in
     * encrypted files header.
     */
    

    private void parseHeader(
        File            file_in,
        SecureRandom    sec_rand,
        PrivateKey      priv_key)
        throws Exception
    {
    
        //
        // Set up input stream wrappers to read the file.
        //

        DataInputStream data_in = new DataInputStream(new FileInputStream(file_in));
        int             l = 0; // Universal length variable.
        boolean         ena = false; // Enable flag.. (Set when file header found)
        boolean         stop = false; // Stop flag.

        Cipher          rsa_eng = Cipher.getInstance("RSA/None/OAEPPadding", "BC"); // RSA Cipher instance.
        rsa_eng.init(Cipher.DECRYPT_MODE, priv_key, sec_rand); // Initialize cipher for decryption.

        while(!stop)
        {           
            try
            {
                int cmd = data_in.readShort(); // Read in block header.
                
                if (cmd == FILE_HEADER) // File header.
                { 
                    ena = true; // Flag file header found. 
                    System.out.println("Header Parse: File Header");
                    continue;
                }
                
                if (cmd == DATA_BLOCK) // Read in data block
                {
                    if (!ena)
                    {
                        throw new Exception("Broken header");
                    }
                    System.out.println("Header Parse: Data Block, size = "+l);
            
                    l = data_in.readInt(); // Read length.
                    data_in.skip(l); // Skip this data.
                    continue;
                }
                
                if (cmd == FINAL_DATA_BLOCK) // Final data block.
                {
                    if (!ena)
                    {
                        throw new Exception("Broken header");
                    }
                    l = data_in.readInt(); // Read length
                    System.out.println("Header Parse: Final Data Block, size = "+l);
                    data_in.skip(l); // Skip this data.
                    continue;
                }
                
                if (cmd == SIG_BLOCK) // Signature block.
                {
                    if (!ena)
                    {
                        throw new Exception("Broken header");
                    }
                    l = data_in.readInt(); // Read length.
                    System.out.println("Parse: Signature block size = "+l);
                    sig_code = new byte[l]; // Create new array (l) in size.
                    data_in.readFully(sig_code); // Read in signature.
                    continue;
                }
                

                if (cmd == KEY_BLOCK) // Read in key block.
                {
                    if (!ena)
                    {
                        throw new Exception("Broken header");
                    }

                    l = data_in.readInt(); // Read length.
                    System.out.println("Parse: Key encoded block size = "+l);
                    byte[] d = new byte[l];  // Create new array (l) in size.
                    data_in.readFully(d); // read in data.
                    
                    //
                    // We must use a SecretKeySpec set up for AES
                    // to convert the raw encoded AES key back into a SecretKey Object.
                    // The key is also decrypted before processing.
                    //

                    aes_key = (SecretKey)new SecretKeySpec(rsa_eng.doFinal(d), "AES");
                    continue;
                }
                
                if (cmd == IV_BLOCK) // Read in IV.
                {
                    if (!ena)
                    {
                        throw new Exception("Broken header");
                    }
                    l = data_in.readInt(); // Read length
                    System.out.println("Parse: IV block, size = "+l);
                    aes_iv = new byte[l]; // Create new array for IV (l) in size.
                    data_in.readFully(aes_iv); // Read in IV.
                    aes_iv = rsa_eng.doFinal(aes_iv); // Decrypt IV
                    continue;
                }
                

                if (cmd == LOCK_BLOCK) // Read lock.
                {
                    if (!ena)
                    {
                        throw new Exception("Broken header");
                    }
    

                    //
                    // Set up cipher to decrypt lock data.
                    //

                    Cipher              dec_lock = Cipher.getInstance("AES/CBC/PKCS7Padding", "BC"); // Get Cipher object.
                    dec_lock.init(Cipher.DECRYPT_MODE, aes_key, new IvParameterSpec(aes_iv));
                    l = data_in.readInt(); // Read Length
                    System.out.println("Parse: Locking data block, size = "+l);
                    byte[] d = new byte[l]; // Create new byte array (l) in size.
                    data_in.readFully(d); // Read lock.
                    lock_data = dec_lock.doFinal(d); // Decrypt lock. 
                    
                    dec_lock = null; // Make null;

                    continue;
                }

                if (cmd == CERT_BLOCK) // Senders certificate block.
                {
                    if (!ena)
                    {
                        throw new Exception("Broken header");
                    }

                    l = data_in.readInt(); // Read Length
                    System.out.println("Parse: Fetch Cert block, size = "+l);

                    byte[]              d = new byte[l]; // Create new array (l) in length.

                    data_in.readFully(d); // Read in certificate.

                    //
                    // The encoded cert must be converted back into an Certificate Object.
                    // More specifically an X509Certificate object.
                    // To do this we use the CertificateFactory set up for our type.
                    // In this case the type is X509.
                    //
                    // See CertificateFactory in the java doc for more info.
                    //

                    CertificateFactory  cf = CertificateFactory.getInstance("X.509"); 
                    their_cert = (X509Certificate) cf.generateCertificate(new ByteArrayInputStream(d));
                    continue;
                }
            }
            catch (EOFException eof)
            {
                stop = true;
            }
        }
        return;
    }

    // Have fun..
}
