package examples.jce.rsa_aes;

import java.io.*;
import javax.crypto.*;
import java.security.*;
import java.security.cert.*;
import javax.crypto.spec.*;


public class EncryptFileSIG
{
    //
    // Define our block header values.
    //

    static final int FILE_HEADER = 0x7e01;  // File header.
    static final int DATA_BLOCK = 1; // Data block
    static final int FINAL_DATA_BLOCK = 2; // Final data block.
    static final int SIG_BLOCK = 3; // Signature block
    static final int CERT_BLOCK = 4; // Certificate block.

    static final int KEY_BLOCK = 16; // Key block
    static final int IV_BLOCK = 17; // IV block
    static final int LOCK_BLOCK = 18; // Locking block

    public static void main(String[] args)
    {
        System.out.println("Encrypt file with and use Signature (MD5withRSA) to provide"); // Print statements.
        System.out.println("sender authentication and message integrity checking"); // More print statements.

        SecureRandom    sec_rand = null; // Secure random number


        //
        // Add BouncyCastle JCE as a provider and create and seed the random number generator.
        //

        try
        {
            Security.addProvider(new org.bouncycastle.jce.provider.BouncyCastleProvider()); // Add provider.
            sec_rand = SecureRandom.getInstance("SHA1PRNG","SUN"); // Create random number generator.

            //
            // Try and seed PRNG from '/dev/urandom' on UNIX machines this device generally exists.
            // if it does not then SecureRandom will employ its own platform independent methods.
            //

            if (new File("/dev/urandom").exists()) // Does device micro random exist.
            {
                System.out.println("Salting SecureRandom (SHA1PRNG) from /dev/urandom");
                byte[] salt = new byte[8192]; // salt.
                new FileInputStream("/dev/urandom").read(salt); // Read salt.
                sec_rand.setSeed(salt); // Set salt.
            }
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }

        //
        // Check from minimum number of command line arguments.
        //

        if (args.length <4)
        {
            System.out.println("Usage: <source file> <dest file> <recipients certificate> <your trust store>");
            System.exit(0);
        }

        File    file_in = new File(args[0]); // Source of plain text.
        File    file_out = new File(args[1]); // Destination of data.

        //
        // Check that source file exists.
        //

        if (!file_in.exists())
        {
            System.out.println("Source File does not exist..!");
            System.exit(0);
        }
    
        //
        // Check that recipients certificate file exists.
        // 

        File    cert_file = new File(args[2]); // Certificate file.

        if (!cert_file.exists()) // Does it _not_ exist.
        {
            System.out.println("Recipients Certificate File does not exist..!");
            System.exit(0);
        }

        //
        // Check that key store file exists. 
        //

        File    keystore_file = new File(args[3]);

        if (!keystore_file.exists()) // Does it _not_ exist.
        {
            System.out.println("Keystore file specified could not be found.");
            System.exit(0);
        }



        char[]          ks_pass = null; // Keystore password.
        boolean         stop = false; // Stop flag.
        KeyStore        key_store = null; // Key Store object.
        PrivateKey      priv_key = null; // Private key.
        X509Certificate my_cert = null; // Certificate.

        try
        {
            key_store = KeyStore.getInstance("JKS", "SUN"); // Create a new instance of the JKS key store.        

            //
            // We need to obtain the password for the key store from the user.
            // Use our getLine() method to read it in.
            // Exit program on zero length input.
            // Keep looping until user inputs a password that works.
            //

            while(!stop)
            {
                System.out.println("");
                System.out.print("Your Key Store Password > ");
                String l = getLine(); // Read line.

                if (l == null) // Null entry.
                {
                    System.exit(0);
                }
                
                if (l.length() == 0) // Zero length.
                {
                    System.exit(0);
                }

                ks_pass = l.toCharArray(); // Need to convert to char[] .

                //
                // Test password on keystore.
                // KeyStore will break when the password is incorrect and
                // throw an exception.
                //

                try
                {
                    key_store.load(new FileInputStream(keystore_file),ks_pass);
                    stop=true; // it worked so stop looping..
                }
                catch (Exception ex)
                {
                    System.out.println(ex.getMessage());
                }
            }
        
            stop = false; // Reset stop flag.
        

            //
            // We now need to get an alias for the private key and our certificate 
            // that we will use to sign the encrypted file so that the recipient
            // can authenticate us.
            // This uses the same process as described above.

            while(!stop)
            {
                System.out.println("");
                System.out.print("Secret Key Alias for Signing only use > ");
                String l = getLine();
                if (l == null)
                {
                    System.exit(0);
                }
                
                if (l.length() == 0)
                {
                    System.exit(0);
                }
                
                try
                {
                    priv_key = (PrivateKey)key_store.getKey(l, ks_pass); // Get private key.
                    my_cert = (X509Certificate)key_store.getCertificate(l); // Get Signing Certificate
                    stop = true;
                }
                catch (Exception ex)
                {
                    System.out.println(ex.getMessage());
                }
            }


            //
            // Now we need to read in the certificate of the recipient.
            // The certificate contains the public key that the recipient uses for encryption.
            // See javadoc for CertificateFactory for more info..
            //
        
            System.out.println("Reading recipients certificate file.");
      
            byte[]          raw = new byte[(int)cert_file.length()]; // Create new array
            DataInputStream din = new DataInputStream(new FileInputStream(cert_file)); // Some stream wrapper.

            din.readFully(raw); // Read it all in.
            din.close(); // Close file input.
    
            ByteArrayInputStream    b_in = new ByteArrayInputStream(raw); // Need to use this..
            
            //
            // Use a certificate factory to convert the raw certificate into a certificate object.
            //

            CertificateFactory      cf = CertificateFactory.getInstance("X.509"); // Create certificate factory.
            X509Certificate         x_cert = (X509Certificate)cf.generateCertificate(b_in); // Create certificate.


            //
            // Now we have all the necessary information to encrypt a file using signatures for authentication
            // we can call the process file method.
            //

            processFile(file_in, x_cert, my_cert, priv_key, sec_rand, file_out);
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }

    /**
     * zero out the passed in byte array
     */
    private static void blank(
        byte[]   bytes)
    {
        for (int t = 0; t < bytes.length; t++)
        {
            bytes[t]=0;
        }
    }

    /**
     * Read in line from System.in.
     */

    private static String getLine()
    {
        try
        {
            BufferedReader bin = new BufferedReader(new InputStreamReader(System.in));
            return(bin.readLine());
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }

        return(null);
    }



    /**
     * Encrypt file.
     */

    public static void processFile(
        File            file_in,
        X509Certificate x_cert,
        X509Certificate my_cert,
        PrivateKey      my_private,
        SecureRandom    sec_rand,
        File            file_out)
        throws Exception
    {
        processFile(file_in,
                x_cert.getPublicKey(), my_cert, my_private, sec_rand, file_out);
    }

    /**
     * Encrypt file..
     */

    public static void processFile(
        File            file_in,
        PublicKey       pub_key, 
        X509Certificate my_cert,
        PrivateKey      my_private,
        SecureRandom    sec_rand,
        File            file_out)
        throws Exception
    {
        sec_rand = SecureRandom.getInstance("SHA1PRNG","SUN");
    
        //
        // Generate key for AES cipher at 128 bits.
        // Using the appropriate key generator.
        // It is important that you use the correct key generator.
        // As it will filter out any keys that may be weak on a cipher by cipher
        // basis.
        //

        System.out.println("Generate key for AES at 128 bits");
        KeyGenerator    key_gen = KeyGenerator.getInstance("AES","BC"); // Get instance of an AES generator.

        key_gen.init(128,sec_rand); // Configure for 128 bits using out previously salted PRNG.

        Key             aes_key = key_gen.generateKey(); //Generate key.
    
        //
        // Instantiate Symmetric cipher for encryption.
        //
        System.out.println("Set up cipher AES/CBC/PKCS7Padding");
        Cipher  output_cipher = Cipher.getInstance("AES/CBC/PKCS7Padding","BC"); // Cipher object. 

        output_cipher.init(Cipher.ENCRYPT_MODE,aes_key,sec_rand); // Initialize cipher.
    
        //
        // Get key and IV for cipher so that they can be later
        // encrypted to build a header.
        //
        byte[] aes_key_enc = aes_key.getEncoded();

        //
        // The cipher automatically generates the IV if none is supplied
        // and one is required.
        // You need to collect the IV.
        //

        byte[] aes_iv = output_cipher.getIV();
    
        //
        // Set up locking data so that one piece of the Signature info or
        // HMAC is based on data that is not directly transmitted. If this
        // is not done the message could be resigned by an attacker if
        // the sender has been compromised and their private key is known.
        // (Which you should expect.)
        //
        // I generally prefer to use locking data like this so that the
        // message can be validated as original without having to be fully
        // decrypted first.
        //
        // As another approach you could take the Signature or HMAC of the
        // plain text of the message the result is still the same as if you
        // had chosen suitably large locking data.
        //
        // NOTE: In the case of an HMAC the locking data is the key to the HMAC.
        //
        byte[]      lock = new byte[24];

        sec_rand.nextBytes(lock);

        //
        // Setup Signature
        //
        System.out.println("Generate lock data to tie signature to file");
    
        Signature   sig = Signature.getInstance("MD5withRSA","BC"); // Set up signature object.

        sig.initSign(my_private,sec_rand); // Initialize with my private signing key.
        sig.update(lock); // put plain text of lock data into signature.
    
    
        //
        // Setup RSA to encrypt secrets.
        //
        System.out.println("Set up RSA with OAEPPadding (see PKCS1 V2)");
        Cipher rsa_eng = Cipher.getInstance("RSA/None/OAEPPadding", "BC"); // Set up cipher for RSA.
        rsa_eng.init(Cipher.ENCRYPT_MODE, pub_key, sec_rand); // Initialize.
    
        //
        // Setup to process File.
        //
        FileInputStream         in_str = new FileInputStream(file_in); // Source of plain text.
        FileOutputStream        file_str = new FileOutputStream(file_out); // Final output stream.
        SignatureOutputStream   sig_str = new SignatureOutputStream(file_str, sig); // As it says.
        DataOutputStream        data_str = new DataOutputStream(sig_str); // See java doc.
    


        //
        // We are now going to form the header for the encrypted file.
        // 

        System.out.println(file_out.getName()+" >> file header");
        data_str.writeShort(FILE_HEADER); // Write a file header.

        //
        // Write out a block for the key of the AES cipher.
        //
        System.out.println(file_out.getName()+" >> Aes key enc with RSA");
        
        data_str.writeShort(KEY_BLOCK); // Block header.
        byte[]  tmp = rsa_eng.doFinal(aes_key_enc); // Encrypt it with RSA.
        data_str.writeInt(tmp.length); // Write length.
        data_str.write(tmp); // Write data.
        blank(tmp); // Erase tmp array.
    
        //
        // This will upset some people because you don't have to keep the
        // IV a secret.
        //
        // I keep the IV secret because I prefer to keep things secret
        // when I can. 
        // 
        System.out.println(file_out.getName()+" >> Aes IF enc with RSA (See note in source code)");
        data_str.writeShort(IV_BLOCK);  // Block header       
        tmp = rsa_eng.doFinal(aes_iv); // Encrypt with RSA.
        data_str.writeInt(tmp.length); // Write length.
        data_str.write(tmp); // Write data.
        blank(tmp); // Clear.
    
        //
        // Write out lock data for SIGNATURE.
        //
        // This system takes the signature on the output of the processing and to prevent it
        // from being resigned and messed about with by attackers we incorperate into the signature
        // some salt that is randomly generated.
        // This salt is encrypted with the Symmetric cipher.
        // This forces the attacker to have to be able to decrypt the AES key and IV before they
        // can alter the message in any way.
        //

        System.out.println(file_out.getName()+" >> Signature lock data enc with RSA");

        data_str.writeShort(LOCK_BLOCK); // Write header.        
        tmp = output_cipher.doFinal(lock); // Encrypt with AES.
        data_str.writeInt(tmp.length); // Write length.
        data_str.write(tmp); // Write data.
        blank(tmp); // Clear.
            

        //
        // Reset AES cipher back to original, as you can not rely on the doFinal() method to do it.
        //
        
        output_cipher.init(Cipher.ENCRYPT_MODE,aes_key,new IvParameterSpec(aes_iv)); // initialize with aes_key.
            


        //
        // Now we must set up to process the file.
        //

        int     l = 0; // Universal length variable.
        byte[]  buf = new byte[8192]; // A buffer to work in.
        byte[]  out = null; // Output buffer.

        //
        // Read while length is > -1
        //

        while((l = in_str.read(buf)) > - 1)
        {
            out = output_cipher.update(buf,0,l); // Encrypt data.
            data_str.writeShort(DATA_BLOCK); // Write data block header.
            data_str.writeInt(out.length); // Write length.
            data_str.write(out); // Write encrypted data.
            System.out.println(file_out.getName()+" >> Encrypted "+out.length+" bytes output");
        }

        //
        // This is the last block
        // It will contain the final output of the cipher
        // which will have any remaining plain text less that the block
        // size of AES (16 bytes) padded PKCS7 style encrypted.
        // Note use of cipher.doFinal();
        //

        out = output_cipher.doFinal(); // Do final encryption.
        data_str.writeShort(FINAL_DATA_BLOCK); // Write header.
        data_str.writeInt(out.length); // Write length.
        data_str.write(out); // Write Data.
        System.out.println(file_out.getName()+" >> Final Encrypted "+out.length+" bytes output");

        blank(buf); // Clear buffer.
        buf = null; // Set Null



        //
        // Now we need to write out our certificate that we
        // use for signing so that the recipient can verify our
        // identity.
        //

        data_str.writeShort(CERT_BLOCK); // Cert block header.
        tmp = my_cert.getEncoded(); // Get encoded in a byte array.
        data_str.writeInt(tmp.length); // Write length.
        data_str.write(tmp); // Write data.

        //
        // The last block we write out is our
        // certificate block.
        //

        System.out.println(file_out.getName()+" >> Write out Signature code block.");
        data_str.writeShort(SIG_BLOCK); // Write Header.
        data_str.flush(); // Flush it..

        //
        // The flush was called because we now need to
        // finalize our signature but calling Signature.sign()
        // so all the data needs to flushed first.
        //

        tmp = sig.sign(); // Get signature code.
        data_str.writeInt(tmp.length); // Write length.
        data_str.write(tmp); // Write data.
        blank(tmp); // Clear.
    
        //
        // Flush and close output.
        //

        data_str.flush();
        data_str.close();
    
        //
        // Get rid of key as best we can.
        //

        System.out.println("Dispose of key material as best we can in java..");
        blank(aes_key_enc);     // Clear encoded key array.
        aes_key_enc = null;     // Set null.
        aes_key = null;         // Best we can do on a Key object; 
        tmp = null;             // Get rid of tmp also.
    
        System.out.println("The End..");
    }
} 
